# pylint: disable=abstract-method
"""
{{ cookiecutter.description }}
"""
from typing import Any, Dict, Iterator, List, Optional, Tuple

import requests_cache

from shillelagh.adapters.base import Adapter
from shillelagh.fields import Field
from shillelagh.filters import Filter
from shillelagh.typing import RequestedOrder, Row


class {{ cookiecutter.adapter_name|replace(' ', '') }}{{ cookiecutter.adapter_type }}(Adapter):

    """
    {{ cookiecutter.description }}
    """

    # Set this to ``True`` if the adapter doesn't access the filesystem.
    safe = False

    # This method is used to determine which URIs your adapter will handle. For
    # example, if your adapter interfaces with an API at api.example.com you
    # can do:
    #
    #     parsed = urllib.parse.urlparse(uri)
    #     return parsed.netloc == "api.example.com"
    #
    # The method will be called 1 or 2 times. On the first call ``fast`` will be
    # true, and the adapter should return as soon as possible. This means that if
    # you need to do a network request to determine if the URI should be handled
    # by your adapter, when ``fast=True`` you should return ``None``.
    #
    # On the second call, ``fast`` will be false, and the adapter can make network
    # requests to introspect URIs. The adapter should then return either true or
    # false on the second request.
    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        return False

    # This method parses the URI into arguments that are passed to initialize the
    # adapter class. The simplest implementation returns the URI unmodified.
    @staticmethod
    def parse_uri(uri: str) -> Tuple[str]:
        return (uri,)

    def __init__(self, uri: str):
        super().__init__()
        self.uri = uri

        # If the adapter needs to do API requests it's useful to use a cache for
        # the requests. If you're not doing network requests you get delete this
        # session object.
        self._session = requests_cache.CachedSession(
            cache_name="{{ cookiecutter.slug }}_cache",
            backend="sqlite",
            expore_after=180,
        )

        # For adapters with dynamic columns (ie, number, names, and types of
        # columns depend on the URI) it's good practice to set the columns when
        # the class is initialized.
        self._set_columns()

    # When defining columns it's important to know which columns can be used to
    # filter the URI. For example, if the API accepts a time range the adapter
    # probably needs a temporal column that is filterable:
    #
    #     from shillelagh.fields import DateTime
    #     from shillelagh.filters import Range
    #
    #     self.columns["time"] = DateTime(filters=[Range])
    #
    # It's then up to the ``get_rows`` method to translate temporal filters into
    # the corresponding API calls.
    #
    # The column definition should also specify if the column has a natural order,
    # or if the adapter can handle any requested order (and the adapter will be
    # responsible for ordering the data):
    #
    #     from shillelagh.fields import Integer
    #     from shillelagh.fields import Order
    #
    #     self.columns["prime_numbers"] = Integer(
    #         filters=[Range],
    #         order=Order.ASCENDING,
    #     )
    #
    # Finally, columns can return data that is not filtered perfectly — eg, hourly
    # data filtered only down to the daily granularity. In that case the column
    # should be declared as inexact:
    #
    #     self.columns["time"] = DateTime(filters=[Range], exact=False)
    #
    def _set_columns(self) -> None:
        self.columns: Dict[str, Field] = {}

    # If you have static columns you can get rid of the ``get_columns`` method,
    # get rid of ``_set_columns``, and simply define the columns as class
    # attributes:
    #
    # time = DateTime(filters=[Range])
    #
    # Then ``get_columns`` from the parent class will find these attributes and
    # use them as the dictionary of columns.
    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    # This method is used to return any extra metadata associated with the URI.
    # You can delete it instead of returning an empty dictionary, since that's
    # the exact implementation of the parent method.
    def get_metadata(self) -> Dict[str, Any]:
        return {}

    # This method yields rows of data, each row a dictionary. If any columns are
    # declared as filterable there might be a corresponding ``Filter`` object in
    # the ``bounds`` argument that must be used to filter the column (unless the
    # column was declared as inexact).
    def get_rows(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> Iterator[Row]:
        pass



"""
Test for shillelagh.backends.apsw.dialects.gsheets.
"""
import datetime
from typing import Any, Dict
from unittest import mock

import pytest
import requests
import sqlalchemy
from pytest_mock import MockerFixture
from requests_mock.mocker import Mocker
from sqlalchemy import MetaData, Table, inspect
from sqlalchemy.engine import create_engine
from sqlalchemy.engine.url import make_url

from shillelagh.backends.apsw.dialects.gsheets import APSWGSheetsDialect, extract_query
from shillelagh.exceptions import ProgrammingError

from ....fakes import incidents


def test_gsheets_dialect() -> None:
    """
    Test the basic usage of the dialect
    """
    dialect = APSWGSheetsDialect()
    assert dialect.create_connect_args(make_url("gsheets://")) == (
        (),
        {
            "path": ":memory:",
            "adapters": ["gsheetsapi"],
            "adapter_kwargs": {
                "gsheetsapi": {
                    "access_token": None,
                    "service_account_file": None,
                    "service_account_info": None,
                    "subject": None,
                    "catalog": {},
                    "app_default_credentials": False,
                },
            },
            "safe": True,
            "isolation_level": None,
        },
    )

    dialect = APSWGSheetsDialect(
        service_account_info={"secret": "XXX"},
        subject="user@example.com",
    )
    assert dialect.create_connect_args(make_url("gsheets://")) == (
        (),
        {
            "path": ":memory:",
            "adapters": ["gsheetsapi"],
            "adapter_kwargs": {
                "gsheetsapi": {
                    "access_token": None,
                    "service_account_file": None,
                    "service_account_info": {"secret": "XXX"},
                    "subject": "user@example.com",
                    "catalog": {},
                    "app_default_credentials": False,
                },
            },
            "safe": True,
            "isolation_level": None,
        },
    )

    dialect = APSWGSheetsDialect(
        service_account_file="credentials.json",
        subject="user@example.com",
        catalog={"public_sheet": "https://example.com/"},
    )
    assert dialect.create_connect_args(make_url("gsheets://")) == (
        (),
        {
            "path": ":memory:",
            "adapters": ["gsheetsapi"],
            "adapter_kwargs": {
                "gsheetsapi": {
                    "access_token": None,
                    "service_account_file": "credentials.json",
                    "service_account_info": None,
                    "subject": "user@example.com",
                    "catalog": {"public_sheet": "https://example.com/"},
                    "app_default_credentials": False,
                },
            },
            "safe": True,
            "isolation_level": None,
        },
    )

    dialect = APSWGSheetsDialect(
        app_default_credentials=True,
    )
    assert dialect.create_connect_args(make_url("gsheets://")) == (
        (),
        {
            "path": ":memory:",
            "adapters": ["gsheetsapi"],
            "adapter_kwargs": {
                "gsheetsapi": {
                    "access_token": None,
                    "service_account_file": None,
                    "service_account_info": None,
                    "subject": None,
                    "catalog": {},
                    "app_default_credentials": True,
                },
            },
            "safe": True,
            "isolation_level": None,
        },
    )

    mock_dbapi_connection = mock.MagicMock()
    assert dialect.get_schema_names(mock_dbapi_connection) == []


def test_get_table_names(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test ``get_table_names``.
    """
    get_credentials = mocker.patch(
        "shillelagh.backends.apsw.dialects.gsheets.get_credentials",
    )

    session = requests.Session()
    session.mount("https://", requests_mock)
    mocker.patch(
        "shillelagh.backends.apsw.dialects.gsheets.AuthorizedSession",
        return_value=session,
    )
    requests_mock.register_uri(
        "GET",
        (
            "https://www.googleapis.com/drive/v3/files?"
            "q=mimeType='application/vnd.google-apps.spreadsheet'"
        ),
        json={"files": [{"id": 1}, {"id": 2}, {"id": 3}]},
    )
    requests_mock.register_uri(
        "GET",
        "https://sheets.googleapis.com/v4/spreadsheets/1?includeGridData=false",
        json={
            "sheets": [
                {"properties": {"sheetId": 0}},
                {"properties": {"sheetId": 1}},
            ],
        },
    )
    requests_mock.register_uri(
        "GET",
        "https://sheets.googleapis.com/v4/spreadsheets/2?includeGridData=false",
        json={"sheets": [{"properties": {"sheetId": 0}}]},
    )
    requests_mock.register_uri(
        "GET",
        "https://sheets.googleapis.com/v4/spreadsheets/3?includeGridData=false",
        json={
            "error": {
                "code": 403,
                "message": (
                    "Google Sheets API has not been used in project 1034909279888 "
                    "before or it is disabled. Enable it by visiting "
                    "https://console.developers.google.com/apis/api/"
                    "sheets.googleapis.com/overview?project=1034909279888 then "
                    "retry. If you enabled this API recently, wait a few minutes "
                    "for the action to propagate to our systems and retry."
                ),
                "status": "PERMISSION_DENIED",
                "details": [
                    {
                        "@type": "type.googleapis.com/google.rpc.Help",
                        "links": [
                            {
                                "description": "Google developers console API activation",
                                "url": (
                                    "https://console.developers.google.com/apis/api/"
                                    "sheets.googleapis.com/overview?project=1034909279888"
                                ),
                            },
                        ],
                    },
                    {
                        "@type": "type.googleapis.com/google.rpc.ErrorInfo",
                        "reason": "SERVICE_DISABLED",
                        "domain": "googleapis.com",
                        "metadata": {
                            "consumer": "projects/1034909279888",
                            "service": "sheets.googleapis.com",
                        },
                    },
                ],
            },
        },
    )
    _logger = mocker.patch("shillelagh.backends.apsw.dialects.gsheets._logger")

    engine = create_engine("gsheets://", list_all_sheets=True)

    get_credentials.return_value = None
    inspector = inspect(engine)
    tables = inspector.get_table_names()
    assert tables == []

    get_credentials.return_value = "SECRET"
    inspector = inspect(engine)
    tables = inspector.get_table_names()
    assert tables == [
        "https://docs.google.com/spreadsheets/d/1/edit#gid=0",
        "https://docs.google.com/spreadsheets/d/1/edit#gid=1",
        "https://docs.google.com/spreadsheets/d/2/edit#gid=0",
    ]

    _logger.warning.assert_called_with(
        "Error loading sheets from file: %s",
        (
            "Google Sheets API has not been used in project 1034909279888 before "
            "or it is disabled. Enable it by visiting "
            "https://console.developers.google.com/apis/api/"
            "sheets.googleapis.com/overview?project=1034909279888 then retry. "
            "If you enabled this API recently, wait a few minutes for the action "
            "to propagate to our systems and retry."
        ),
    )


def test_drive_api_disabled(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test error message when the Drive API is disabled.
    """
    get_credentials = mocker.patch(
        "shillelagh.backends.apsw.dialects.gsheets.get_credentials",
    )

    session = requests.Session()
    session.mount("https://", requests_mock)
    mocker.patch(
        "shillelagh.backends.apsw.dialects.gsheets.AuthorizedSession",
        return_value=session,
    )
    requests_mock.register_uri(
        "GET",
        (
            "https://www.googleapis.com/drive/v3/files?"
            "q=mimeType='application/vnd.google-apps.spreadsheet'"
        ),
        json={
            "error": {
                "errors": [
                    {
                        "domain": "usageLimits",
                        "reason": "accessNotConfigured",
                        "message": (
                            "Access Not Configured. Drive API has not been used in "
                            "project 1034909279888 before or it is disabled. Enable "
                            "it by visiting https://console.developers.google.com/apis/api/"
                            "drive.googleapis.com/overview?project=1034909279888 then retry. "
                            "If you enabled this API recently, wait a few minutes for the "
                            "action to propagate to our systems and retry."
                        ),
                        "extendedHelp": (
                            "https://console.developers.google.com/apis/api/"
                            "drive.googleapis.com/overview?project=1034909279888"
                        ),
                    },
                ],
                "code": 403,
                "message": (
                    "Access Not Configured. Drive API has not been used in project "
                    "1034909279888 before or it is disabled. Enable it by visiting "
                    "https://console.developers.google.com/apis/api/"
                    "drive.googleapis.com/overview?project=1034909279888 then retry. "
                    "If you enabled this API recently, wait a few minutes for the "
                    "action to propagate to our systems and retry."
                ),
            },
        },
    )

    engine = create_engine("gsheets://", list_all_sheets=True)

    get_credentials.return_value = "SECRET"
    inspector = inspect(engine)
    with pytest.raises(ProgrammingError) as excinfo:
        inspector.get_table_names()

    assert str(excinfo.value) == (
        "Access Not Configured. Drive API has not been used in project 1034909279888 "
        "before or it is disabled. Enable it by visiting "
        "https://console.developers.google.com/apis/api/"
        "drive.googleapis.com/overview?project=1034909279888 then retry. If you "
        "enabled this API recently, wait a few minutes for the action to propagate "
        "to our systems and retry."
    )


def test_extract_query() -> None:
    """
    Test ``extract_query``.

    Older version of SQLAlchemy (<1.4) have a bug in ``make_url``.
    """
    assert extract_query(make_url("gsheets://")) == {}
    assert extract_query(make_url("gsheets://host")) == {}
    assert extract_query(make_url("gsheets://?foo=bar")) == {"foo": "bar"}
    assert extract_query(make_url("gsheets:///?foo=bar")) == {"foo": "bar"}
    assert extract_query(make_url("gsheets://?app_default_credentials=TRUE")) == {
        "app_default_credentials": True,
    }
    assert extract_query(make_url("gsheets://?app_default_credentials=1")) == {
        "app_default_credentials": True,
    }
    assert extract_query(make_url("gsheets://?app_default_credentials=True")) == {
        "app_default_credentials": True,
    }
    assert extract_query(make_url("gsheets://?app_default_credentials=true")) == {
        "app_default_credentials": True,
    }


def test_do_ping(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test ``do_ping``.
    """
    connection = mocker.MagicMock()

    dialect = APSWGSheetsDialect()

    requests_mock.get(
        "https://www.google.com/appsstatus/dashboard/incidents.json",
        json=incidents,
    )
    assert dialect.do_ping(connection)

    outage = {
        "service_name": "Google Sheets",
        "modified": "2021-04-12T17:00:00+00:00",
        "most_recent_update": {"status": "SERVICE_OUTAGE"},
    }
    requests_mock.get(
        "https://www.google.com/appsstatus/dashboard/incidents.json",
        json=[outage],
    )
    assert dialect.do_ping(connection) is False

    requests_mock.get(
        "https://www.google.com/appsstatus/dashboard/incidents.json",
        json=[],
    )
    assert dialect.do_ping(connection)


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
def test_types_in_sqlalchemy(adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test that the SQLAlchemy dialect returns types correctly.

    The Shillelagh dialect is based on the SQLite dialect, but there's a major difference
    between them: in SQLite date/time/datetime values are returned as strings, while in
    Shillelagh they are returned as Python objects.
    """
    engine = create_engine("gsheets://", **adapter_kwargs["gsheetsapi"])
    metadata = MetaData()
    metadata.reflect(engine)
    table = Table(
        "https://docs.google.com/spreadsheets/d/"
        "1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit#gid=1648320094",
        metadata,
        autoload_with=engine,
    )

    assert [(column.name, column.type.__class__) for column in table.c] == [
        ("datetime", sqlalchemy.types.TIMESTAMP),
        ("number", sqlalchemy.types.REAL),
        ("boolean", sqlalchemy.types.BOOLEAN),
        ("date", sqlalchemy.types.DATE),
        ("timeofday", sqlalchemy.types.TIME),
        ("string", sqlalchemy.types.TEXT),
    ]

    connection = engine.connect()
    results = connection.execute(table.select()).fetchall()
    assert list(results) == [
        (
            datetime.datetime(2018, 9, 1, 7, 0, tzinfo=datetime.timezone.utc),
            1.0,
            True,
            datetime.date(2018, 1, 1),
            datetime.time(17, 0),
            "test",
        ),
        (
            datetime.datetime(2018, 9, 2, 7, 0, tzinfo=datetime.timezone.utc),
            1.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 3, 7, 0, tzinfo=datetime.timezone.utc),
            2.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 4, 7, 0, tzinfo=datetime.timezone.utc),
            3.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 5, 7, 0, tzinfo=datetime.timezone.utc),
            5.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 6, 7, 0, tzinfo=datetime.timezone.utc),
            8.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 7, 7, 0, tzinfo=datetime.timezone.utc),
            13.0,
            False,
            None,
            None,
            None,
        ),
        (
            datetime.datetime(2018, 9, 8, 7, 0, tzinfo=datetime.timezone.utc),
            None,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 9, 7, 0, tzinfo=datetime.timezone.utc),
            34.0,
            None,
            None,
            None,
            "test",
        ),
    ]



"""
A simple example showing the Datasette adapter.
"""
from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(":memory:")
    cursor = connection.cursor()

    SQL = """
SELECT "Latitude" AS "Latitude",
       "Longitude" AS "Longitude",
       COUNT(*) AS count
FROM "https://san-francisco.datasettes.com/sf-trees/Street_Tree_List"
WHERE "Latitude" IS NOT NULL
  AND "Longitude" IS NOT NULL
  AND "qSpecies" = 6
GROUP BY "Latitude",
         "Longitude"
ORDER BY count DESC
LIMIT 1000
OFFSET 0;
    """
    for row in cursor.execute(SQL):
        print(row)



"""
Test the base parser/tokenizer.
"""
# pylint: disable=protected-access
from datetime import datetime

import pytest

from shillelagh.adapters.api.gsheets.parsing.base import (
    LITERAL,
    InvalidValue,
    is_unescaped_literal,
    tokenize,
)
from shillelagh.adapters.api.gsheets.parsing.date import DD, MM, YYYY


def test_literal_token():
    """
    Test the literal token.
    """
    classes = [
        DD,
        MM,
        YYYY,
        LITERAL,
    ]

    assert LITERAL.match(r"\d")
    assert LITERAL.match('"dd/mm/yy"')
    # matches eveything
    assert LITERAL.match("d")

    token = LITERAL("@")
    tokens = list(tokenize("@", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "@"

    token = LITERAL('"dd/mm/yy"')
    tokens = list(tokenize('"dd/mm/yy"', classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "dd/mm/yy"

    token = LITERAL('"invalid"')
    assert token.parse("invalid", tokens) == ({}, "")

    token = LITERAL(r"\d")
    assert token.parse("d", tokens) == ({}, "")

    token = LITERAL('"test"')
    with pytest.raises(InvalidValue) as excinfo:
        token.parse("invalid", tokens)
    assert str(excinfo.value) == "invalid"

    token = LITERAL("A")
    with pytest.raises(InvalidValue) as excinfo:
        token.parse("B", tokens)
    assert str(excinfo.value) == "B"

    token = LITERAL(r"\.")
    with pytest.raises(InvalidValue) as excinfo:
        token.parse("B", tokens)
    assert str(excinfo.value) == "B"


def test_tokenize():
    """
    Test the tokenize function.
    """
    classes = [
        DD,
        MM,
        YYYY,
        LITERAL,
    ]
    tokens = list(tokenize('dd/mm/yyyy -> ("dd/mm/yyyy")', classes))
    assert tokens == [
        DD("dd"),
        LITERAL("/"),
        MM("mm"),
        LITERAL("/"),
        YYYY("yyyy"),
        LITERAL(" -> ("),
        LITERAL('"dd/mm/yyyy"'),
        LITERAL(")"),
    ]


def test_is_unescaped_literal():
    """
    Test the is_unescaped_literal function.
    """
    assert is_unescaped_literal(LITERAL("a"))
    assert not is_unescaped_literal(LITERAL(r"\d"))
    assert not is_unescaped_literal(LITERAL('"hello"'))
    assert not is_unescaped_literal(YYYY("yyyy"))



# pylint: disable=abstract-method
"""
A dialect that only connects to GSheets.

This dialect was implemented to replace the ``gsheetsdb`` library.
"""
import logging
import urllib.parse
from datetime import timedelta
from operator import itemgetter
from typing import Any, Dict, List, Optional, Tuple, cast

import requests
from google.auth.transport.requests import AuthorizedSession
from sqlalchemy.engine.url import URL
from sqlalchemy.pool.base import _ConnectionFairy
from typing_extensions import TypedDict

from shillelagh.adapters.api.gsheets.lib import get_credentials
from shillelagh.backends.apsw.dialects.base import APSWDialect
from shillelagh.exceptions import ProgrammingError

_logger = logging.getLogger(__name__)


DEFAULT_TIMEOUT = timedelta(minutes=3)


class QueryType(TypedDict, total=False):
    """
    Types for parameters in the SQLAlchemy URI query.
    """

    access_token: str
    service_account_file: str
    subject: str
    app_default_credentials: bool


def extract_query(url: URL) -> QueryType:
    """
    Extract the query from the SQLAlchemy URL.
    """
    if url.query:
        parameters = dict(url.query)
    # there's a bug in how SQLAlchemy <1.4 handles URLs without hosts,
    # putting the query string as the host; handle that case here
    elif url.host and url.host.startswith("?"):
        parameters = dict(urllib.parse.parse_qsl(url.host[1:]))  # pragma: no cover
    else:
        parameters = {}

    if "app_default_credentials" in parameters:
        parameters["app_default_credentials"] = parameters[
            "app_default_credentials"
        ].lower() in {"1", "true"}

    return cast(QueryType, parameters)


class APSWGSheetsDialect(APSWDialect):
    """
    Drop-in replacement for gsheetsdb.

    This dialect loads only the "gsheetsapi" adapter. To use it:

        >>> from sqlalchemy.engine import create_engine
        >>> engine = create_engine("gsheets://")

    """

    # This is supported in ``SQLiteDialect``, and equally supported here. See
    # https://docs.sqlalchemy.org/en/14/core/connections.html#caching-for-third-party-dialects
    # for more context.
    supports_statement_cache = True

    name = "gsheets"

    def __init__(  # pylint: disable=too-many-arguments
        self,
        access_token: Optional[str] = None,
        service_account_file: Optional[str] = None,
        service_account_info: Optional[Dict[str, Any]] = None,
        subject: Optional[str] = None,
        catalog: Optional[Dict[str, str]] = None,
        list_all_sheets: bool = False,
        app_default_credentials: bool = False,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        self.access_token = access_token
        self.service_account_file = service_account_file
        self.service_account_info = service_account_info
        self.subject = subject
        self.catalog = catalog or {}
        self.list_all_sheets = list_all_sheets
        self.app_default_credentials = app_default_credentials

    def create_connect_args(self, url: URL) -> Tuple[Tuple[()], Dict[str, Any]]:
        adapter_kwargs: Dict[str, Any] = {
            "access_token": self.access_token,
            "service_account_file": self.service_account_file,
            "service_account_info": self.service_account_info,
            "subject": self.subject,
            "catalog": self.catalog,
            "app_default_credentials": self.app_default_credentials,
        }
        # parameters can be overridden via the query in the URL
        adapter_kwargs.update(extract_query(url))

        return (), {
            "path": ":memory:",
            "adapters": ["gsheetsapi"],
            "adapter_kwargs": {"gsheetsapi": adapter_kwargs},
            "safe": True,
            "isolation_level": self.isolation_level,
        }

    def do_ping(self, dbapi_connection: _ConnectionFairy) -> bool:
        """
        Return Google Sheets API status.
        """
        response = requests.get(
            "https://www.google.com/appsstatus/dashboard/incidents.json",
            timeout=DEFAULT_TIMEOUT.total_seconds(),
        )
        payload = response.json()

        updates = [
            update for update in payload if update["service_name"] == "Google Sheets"
        ]
        if not updates:
            return True

        updates.sort(key=itemgetter("modified"), reverse=True)
        latest_update = updates[0]
        status: str = latest_update["most_recent_update"]["status"]

        if status in {"AVAILABLE", "SERVICE_DISRUPTION"}:
            return True

        # in case we don't understand the status, return True to be conservative
        return status != "SERVICE_OUTAGE"

    def get_table_names(  # pylint: disable=unused-argument
        self,
        connection: _ConnectionFairy,
        schema: str = None,
        sqlite_include_internal: bool = False,
        **kwargs: Any,
    ) -> List[str]:
        """
        Return a list of table names.

        This will query for the authenticated user's spreadsheets, and return
        the URL of each sheet in all the spreadsheets. It's also possible to
        specify a "catalog" of URLs, which are also returned using their short
        names.
        """
        table_names = list(self.catalog.keys())

        query = extract_query(connection.url) if hasattr(connection, "url") else {}
        credentials = get_credentials(
            query.get("access_token", self.access_token),
            query.get("service_account_file", self.service_account_file),
            self.service_account_info,
            query.get("subject", self.subject),
            query.get("app_default_credentials", self.app_default_credentials),
        )
        if not (credentials and self.list_all_sheets):
            return table_names

        session = AuthorizedSession(credentials)

        spreadsheet_ids = get_spreadsheet_ids(session)
        for spreadsheet_id in spreadsheet_ids:
            table_names.extend(get_sheet_urls(spreadsheet_id, session))

        return table_names


def get_spreadsheet_ids(session: AuthorizedSession) -> List[str]:
    """
    Return the ID of all spreadsheets that the user has access to.
    """
    url = (
        "https://www.googleapis.com/drive/v3/files?"
        "q=mimeType='application/vnd.google-apps.spreadsheet'"
    )
    _logger.info("GET %s", url)
    response = session.get(url)
    payload = response.json()
    _logger.debug(payload)
    if "error" in payload:
        raise ProgrammingError(payload["error"]["message"])

    return [file["id"] for file in payload["files"]]


def get_sheet_urls(spreadsheet_id: str, session: AuthorizedSession) -> List[str]:
    """
    Return the URL for all sheets in a given spreadsheet.
    """
    response = session.get(
        "https://sheets.googleapis.com/v4/spreadsheets/"
        f"{spreadsheet_id}?includeGridData=false",
    )
    payload = response.json()
    if "error" in payload:
        _logger.warning(
            "Error loading sheets from file: %s",
            payload["error"]["message"],
        )
        return []

    sheet_urls: List[str] = []
    sheets = payload["sheets"]
    for sheet in sheets:
        sheet_id = sheet["properties"]["sheetId"]
        sheet_urls.append(
            "https://docs.google.com/spreadsheets/d/"
            f"{spreadsheet_id}/edit#gid={sheet_id}",
        )

    return sheet_urls



"""
An adapter for retrieving information on running processes and system utilization (CPU,
memory, disks, network, sensors).

See https://github.com/giampaolo/psutil for more information.
"""
import logging
import time
import urllib.parse
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, Iterator, List, Optional, Set, Tuple, Type, Union

import psutil

from shillelagh.adapters.base import Adapter
from shillelagh.fields import DateTime, Field, Float, Integer, Order
from shillelagh.filters import Filter
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

AVERAGE_NUMBER_OF_ROWS = 100


MEMORY_COLUMNS: Dict[str, Type[Field]] = {
    "total": Integer,
    "available": Integer,
    "percent": Float,
    "used": Integer,
    "free": Integer,
    "active": Integer,
    "inactive": Integer,
    "wired": Integer,
}
SWAP_COLUMNS: Dict[str, Type[Field]] = {
    "total": Integer,
    "used": Integer,
    "free": Integer,
    "percent": Float,
    "sin": Integer,
    "sout": Integer,
}


class ResourceType(str, Enum):
    """
    The type of resource to retrieve.
    """

    CPU = "cpu"
    MEMORY = "memory"
    SWAP = "swap"
    ALL = "all"


def get_columns(resource: ResourceType) -> Dict[str, Field]:
    """
    Build columns depending on the chosen resource.
    """
    memory_prefix = "virtual_" if resource == ResourceType.ALL else ""
    swap_prefix = "swap_" if resource == ResourceType.ALL else ""

    columns: Dict[str, Field] = {}

    if resource in {ResourceType.CPU, ResourceType.ALL}:
        columns.update(
            {
                f"cpu{i}": Float(
                    filters=None,
                    order=Order.NONE,
                    exact=False,
                )
                for i in range(psutil.cpu_count())
            },
        )
    if resource in {ResourceType.MEMORY, ResourceType.ALL}:
        columns.update(
            {
                f"{memory_prefix}{column}": field(
                    filters=None,
                    order=Order.NONE,
                    exact=False,
                )
                for column, field in MEMORY_COLUMNS.items()
            },
        )
    if resource in {ResourceType.SWAP, ResourceType.ALL}:
        columns.update(
            {
                f"{swap_prefix}{column}": field(
                    filters=None,
                    order=Order.NONE,
                    exact=False,
                )
                for column, field in SWAP_COLUMNS.items()
            },
        )

    return columns


class SystemAPI(Adapter):

    """
    An adapter for retrieving system information.
    """

    safe = False

    supports_limit = True
    supports_offset = True
    supports_requested_columns = True

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)
        return parsed.scheme == "system" and (
            # pylint: disable=protected-access
            parsed.netloc in ResourceType._value2member_map_
            or parsed.netloc == ""
        )

    @staticmethod
    def parse_uri(uri: str) -> Union[Tuple[str], Tuple[str, float]]:
        parsed = urllib.parse.urlparse(uri)
        resource = parsed.netloc or "all"
        query_string = urllib.parse.parse_qs(parsed.query)

        if "interval" in query_string:
            return (resource, float(query_string["interval"][0]))
        return (resource,)

    def __init__(self, resource: str, interval: float = 1.0):
        super().__init__()

        self.resource = ResourceType(resource)
        self.interval = interval

        self._set_columns()

    def _set_columns(self) -> None:
        self.columns: Dict[str, Field] = {
            "timestamp": DateTime(filters=None, order=Order.ASCENDING, exact=False),
        }
        self.columns.update(get_columns(self.resource))

    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    def get_data(  # pylint: disable=too-many-arguments
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        requested_columns: Optional[Set[str]] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        requested_columns = requested_columns or set(self.columns.keys())

        rowid = 0
        while limit is None or rowid < limit:
            if offset is not None:
                time.sleep(self.interval * offset)

            row: Dict[str, Any] = {"rowid": rowid}
            if "timestamp" in requested_columns:
                row["timestamp"] = datetime.now(timezone.utc)

            needs_sleep = True

            # add CPU?
            if any(column.startswith("cpu") for column in requested_columns):
                try:
                    row.update(
                        {
                            f"cpu{i}": value / 100.0
                            for i, value in enumerate(
                                psutil.cpu_percent(interval=self.interval, percpu=True),
                            )
                            if f"cpu{i}" in requested_columns
                        },
                    )
                    needs_sleep = False
                except KeyboardInterrupt:
                    return

            # virtual memory
            if self.resource == ResourceType.ALL and any(
                column.startswith("virtual_") for column in requested_columns
            ):
                row.update(
                    {
                        f"virtual_{k}": v
                        for k, v in psutil.virtual_memory()._asdict().items()
                        if f"virtual_{k}" in requested_columns
                    },
                )
            elif self.resource == ResourceType.MEMORY and any(
                column in MEMORY_COLUMNS for column in requested_columns
            ):
                row.update(
                    {
                        k: v
                        for k, v in psutil.virtual_memory()._asdict().items()
                        if k in requested_columns
                    },
                )

            # swap
            if self.resource == ResourceType.ALL and any(
                column.startswith("swap_") for column in requested_columns
            ):
                row.update(
                    {
                        f"swap_{k}": v
                        for k, v in psutil.swap_memory()._asdict().items()
                        if f"swap_{k}" in requested_columns
                    },
                )
            elif self.resource == ResourceType.SWAP and any(
                column in SWAP_COLUMNS for column in requested_columns
            ):
                row.update(
                    {
                        k: v
                        for k, v in psutil.swap_memory()._asdict().items()
                        if k in requested_columns
                    },
                )

            _logger.debug(row)
            yield row
            rowid += 1

            if needs_sleep:
                time.sleep(self.interval)



"""
Tests for the adapter registry.
"""

import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.file.csvfile import CSVFile
from shillelagh.adapters.registry import AdapterLoader
from shillelagh.exceptions import InterfaceError

from ..fakes import FakeAdapter


def test_registry(registry: AdapterLoader) -> None:
    """
    Basic tests for the adapter registry.
    """
    registry.clear()
    assert registry.loaders == {}

    registry.add("dummy", FakeAdapter)
    adapter = registry.load("dummy")
    assert adapter == FakeAdapter


def test_load_error(registry: AdapterLoader) -> None:
    """
    Test errors when loading an adapter.
    """
    registry.clear()

    def load_error() -> None:
        raise ImportError("Error!")

    registry.loaders["dummy"].append(load_error)
    with pytest.raises(InterfaceError) as excinfo:
        registry.load("dummy")
    assert str(excinfo.value) == "Unable to load adapter dummy"

    registry.add("dummy", FakeAdapter)
    assert len(registry.loaders["dummy"]) == 2
    adapter = registry.load("dummy")
    assert adapter == FakeAdapter


def test_register(registry: AdapterLoader) -> None:
    """
    Test that we can register new adapters.
    """
    registry.clear()

    registry.register("csvfile", "shillelagh.adapters.file.csvfile", "CSVFile")
    adapter = registry.load("csvfile")
    assert adapter == CSVFile

    registry.register("invalid", "shillelagh.adapters.file.csvfile", "WhatFile")
    with pytest.raises(InterfaceError) as excinfo:
        registry.load("invalid")
    assert str(excinfo.value) == "Unable to load adapter invalid"


def test_load_only_requested_adapters(registry: AdapterLoader) -> None:
    """
    Test that we only try to load requested adapters.
    """
    registry.clear()

    def load_error() -> None:
        raise ImportError("Error!")

    registry.add("valid", FakeAdapter)
    registry.loaders["invalid"].append(load_error)

    assert registry.load_all(["valid"]) == {"valid": FakeAdapter}
    with pytest.raises(InterfaceError) as excinfo:
        registry.load_all(["valid", "invalid"])
    assert str(excinfo.value) == "Unable to load adapter invalid"
    assert registry.load_all() == {"valid": FakeAdapter}


def test_load_warning(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test that warnings are only logged in safe mode.
    """
    _logger = mocker.patch("shillelagh.adapters.registry._logger")

    def load_error() -> None:
        raise ImportError("Error!")

    registry.loaders["dummy"].append(load_error)

    with pytest.raises(InterfaceError):
        registry.load("dummy")
    assert _logger.warning.not_called()

    with pytest.raises(InterfaceError):
        registry.load("dummy", warn=True)
    assert _logger.warning.called_with("Couldn't load adapter %s", "dummy")



"""
Custom functions available to the SQL backend.
"""
import json
import time
from typing import Any, Dict, List, Type

from importlib_metadata import distribution

from shillelagh.adapters.base import Adapter
from shillelagh.lib import find_adapter

__all__ = ["sleep", "get_metadata", "version"]


def sleep(seconds: int) -> None:
    """
    Sleep for ``n`` seconds.

    This is useful for troubleshooting timeouts::

        sql> SELECT sleep(60);

    """
    time.sleep(seconds)


def get_metadata(
    adapter_kwargs: Dict[str, Dict[str, Any]],
    adapters: List[Type[Adapter]],
    uri: str,
) -> str:
    """
    Return metadata about a given table.

    Returns the name of the adapter that supports the table, as well
    as any extra metadata provided by the adapter::

        sql> SELECT GET_METADATA("https://docs.google.com/spreadsheets/d/1/edit#gid=0");
        GET_METADATA("https://docs.google.com/spreadsheets/d/1/edit#gid=0")
        -------------------------------------------------------------------
        {
            "extra": {
                "Spreadsheet title": "A title",
                "Sheet title": "Another title"
            },
            "adapter": "GSheetsAPI"
        }

    """
    adapter, args, kwargs = find_adapter(uri, adapter_kwargs, adapters)
    instance = adapter(*args, **kwargs)

    return json.dumps(
        {
            "extra": instance.get_metadata(),
            "adapter": adapter.__name__,
        },
    )


def version() -> str:
    """
    Return the current version of Shillelagh.

    As an example::

        sql> SELECT VERSION();
        VERSION()
        -----------
        0.7.4

    """
    return str(distribution("shillelagh").version)



"""
Test the Pandas in-memory adapter.
"""
import pandas as pd
import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.memory.pandas import PandasMemory, find_dataframe
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Order
from shillelagh.filters import Equal, IsNotNull, IsNull, NotEqual, Operator


def test_pandas() -> None:
    """
    Test basic operations with a dataframe.
    """
    mydf = pd.DataFrame(  # noqa: F841  pylint: disable=unused-variable
        [
            {"index": 10, "temperature": 15.2, "site": "Diamond_St"},
            {"index": 11, "temperature": 13.1, "site": "Blacktail_Loop"},
            {"index": 12, "temperature": 13.3, "site": "Platinum_St"},
            {"index": 13, "temperature": 12.1, "site": "Kodiak_Trail"},
        ],
    )

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = "SELECT * FROM mydf"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]

    sql = "SELECT * FROM mydf WHERE temperature > 10 AND temperature < 10"
    cursor.execute(sql)
    assert cursor.fetchall() == []

    sql = "SELECT * FROM mydf WHERE temperature > 13 ORDER BY site DESC"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (12, 13.3, "Platinum_St"),
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
    ]

    sql = "SELECT * FROM mydf WHERE temperature < 13"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (13, 12.1, "Kodiak_Trail"),
    ]

    sql = "SELECT * FROM mydf WHERE site = 'Platinum_St'"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (12, 13.3, "Platinum_St"),
    ]

    sql = """INSERT INTO mydf ("index", temperature, site) VALUES (14, 10.1, 'New_Site')"""
    cursor.execute(sql)
    sql = 'SELECT * FROM mydf WHERE "index" > 11'
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
        (14, 10.1, "New_Site"),
    ]

    sql = "DELETE FROM mydf WHERE site = 'New_Site'"
    cursor.execute(sql)
    sql = 'SELECT * FROM mydf WHERE "index" > 11'
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]

    sql = "UPDATE mydf SET temperature = temperature * 2"
    cursor.execute(sql)
    sql = "SELECT * FROM mydf"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (10, 30.4, "Diamond_St"),
        (11, 26.2, "Blacktail_Loop"),
        (12, 26.6, "Platinum_St"),
        (13, 24.2, "Kodiak_Trail"),
    ]

    sql = "UPDATE mydf SET temperature = temperature / 2"
    cursor.execute(sql)


def test_adapter(mocker: MockerFixture) -> None:
    """
    Test additional operations on the adapter.
    """
    mydf = pd.DataFrame(
        [
            {"index": 10, "temperature": 15.2, "site": "Diamond_St"},
            {"index": 11, "temperature": 13.1, "site": "Blacktail_Loop"},
            {"index": 12, "temperature": 13.3, "site": "Platinum_St"},
            {"index": 13, "temperature": 12.1, "site": "Kodiak_Trail"},
        ],
    )

    mock_find_dataframe = mocker.patch(
        "shillelagh.adapters.memory.pandas.find_dataframe",
    )
    mock_find_dataframe.return_value = mydf

    adapter = PandasMemory("mydf")

    assert list(adapter.get_data({"index": Equal(10)}, [])) == [
        {"rowid": 0, "index": 10, "site": "Diamond_St", "temperature": 15.2},
    ]

    assert list(adapter.get_data({"index": NotEqual(10)}, [])) == [
        {"rowid": 1, "index": 11, "temperature": 13.1, "site": "Blacktail_Loop"},
        {"rowid": 2, "index": 12, "temperature": 13.3, "site": "Platinum_St"},
        {"rowid": 3, "index": 13, "temperature": 12.1, "site": "Kodiak_Trail"},
    ]

    adapter.update_data(
        3,
        {"rowid": 5, "index": 13, "temperature": 12.1, "site": "Kodiak_Trail"},
    )
    assert list(adapter.get_data({}, [])) == [
        {"index": 10, "rowid": 0, "site": "Diamond_St", "temperature": 15.2},
        {"index": 11, "rowid": 1, "site": "Blacktail_Loop", "temperature": 13.1},
        {"index": 12, "rowid": 2, "site": "Platinum_St", "temperature": 13.3},
        {"index": 13, "rowid": 5, "site": "Kodiak_Trail", "temperature": 12.1},
    ]

    adapter.insert_data(
        {"rowid": 7, "index": 14, "temperature": 10.1, "site": "New_Site"},
    )
    assert list(adapter.get_data({}, [])) == [
        {"index": 10, "rowid": 0, "site": "Diamond_St", "temperature": 15.2},
        {"index": 11, "rowid": 1, "site": "Blacktail_Loop", "temperature": 13.1},
        {"index": 12, "rowid": 2, "site": "Platinum_St", "temperature": 13.3},
        {"index": 13, "rowid": 5, "site": "Kodiak_Trail", "temperature": 12.1},
        {"index": 14, "rowid": 7, "site": "New_Site", "temperature": 10.1},
    ]

    adapter.update_data(
        5,
        {"rowid": 3, "index": 13, "temperature": 12.1, "site": "Kodiak_Trail"},
    )
    adapter.delete_data(7)
    assert list(adapter.get_data({}, [])) == [
        {"index": 10, "rowid": 0, "site": "Diamond_St", "temperature": 15.2},
        {"index": 11, "rowid": 1, "site": "Blacktail_Loop", "temperature": 13.1},
        {"index": 12, "rowid": 2, "site": "Platinum_St", "temperature": 13.3},
        {"index": 13, "rowid": 3, "site": "Kodiak_Trail", "temperature": 12.1},
    ]

    with pytest.raises(ProgrammingError) as excinfo:
        list(adapter.get_data({"a": [1, 2, 3]}, []))  # type: ignore
    assert str(excinfo.value) == "Invalid filter: [1, 2, 3]"


def test_adapter_nulls(mocker: MockerFixture) -> None:
    """
    Test operations with nulls on the adapter.
    """
    mydf = pd.DataFrame([{"a": None, "b": 10}, {"a": 20, "b": None}])

    mock_find_dataframe = mocker.patch(
        "shillelagh.adapters.memory.pandas.find_dataframe",
    )
    mock_find_dataframe.return_value = mydf

    adapter = PandasMemory("mydf")

    rows = list(adapter.get_data({"a": IsNull()}, []))
    assert len(rows) == 1
    row = rows[0]
    assert row["a"] != row["a"]  # NaN
    assert row["b"] == 10.0
    assert row["rowid"] == 0

    rows = list(adapter.get_data({"a": IsNotNull()}, []))
    assert len(rows) == 1
    row = rows[0]
    assert row["a"] == 20.0
    assert row["b"] != row["b"]  # NaN
    assert row["rowid"] == 1


outer_df = pd.DataFrame()


def test_find_dataframe() -> None:
    """
    Test that we can find the dataframe.
    """
    mydf = pd.DataFrame()

    def inner_scope() -> pd.DataFrame:
        return find_dataframe("mydf")

    assert inner_scope() is mydf

    assert find_dataframe("outer_df") is outer_df

    assert find_dataframe("foobar") is None
    with pytest.raises(ProgrammingError) as excinfo:
        PandasMemory("foobar")
    assert str(excinfo.value) == "Could not find dataframe"


def test_get_cost(mocker: MockerFixture) -> None:
    """
    Test cost estimation.
    """
    mydf = pd.DataFrame(
        [
            {"index": 10, "temperature": 15.2, "site": "Diamond_St"},
            {"index": 11, "temperature": 13.1, "site": "Blacktail_Loop"},
            {"index": 12, "temperature": 13.3, "site": "Platinum_St"},
            {"index": 13, "temperature": 12.1, "site": "Kodiak_Trail"},
        ],
    )

    mock_find_dataframe = mocker.patch(
        "shillelagh.adapters.memory.pandas.find_dataframe",
    )
    mock_find_dataframe.return_value = mydf

    adapter = PandasMemory(mydf)
    assert adapter.get_cost([], []) == 0

    # linear filtering cost
    assert adapter.get_cost([("one", Operator.EQ)], []) == 1000
    assert adapter.get_cost([("one", Operator.EQ), ("two", Operator.EQ)], []) == 2000

    # linear sorting cost
    assert (
        adapter.get_cost(
            [("one", Operator.EQ), ("two", Operator.EQ)],
            [("one", Order.ASCENDING)],
        )
        == 11965
    )
    assert (
        adapter.get_cost(
            [("one", Operator.EQ), ("two", Operator.EQ)],
            [("one", Order.ASCENDING), ("two", Order.DESCENDING)],
        )
        == 21931
    )


def test_integer_column_names() -> None:
    """
    Test dataframes with column names that are integers.
    """
    mydf = pd.DataFrame(  # noqa: F841  pylint: disable=unused-variable
        [
            [10, 15.2, "Diamond_St"],
            [11, 13.1, "Blacktail_Loop"],
            [12, 13.3, "Platinum_St"],
            [13, 12.1, "Kodiak_Trail"],
        ],
    )

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = 'SELECT "0", "1", "2" FROM mydf'
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]


def test_empty_dataframe() -> None:
    """
    Test that empty dataframes work.
    """
    emptydf = pd.DataFrame({"a": []})  # noqa: F841  pylint: disable=unused-variable

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = "SELECT * FROM emptydf"
    cursor.execute(sql)
    assert cursor.fetchall() == []



# pylint: disable=invalid-name, unused-argument
"""
Tests for shillelagh.console.
"""
from io import StringIO
from pathlib import Path

import yaml
from appdirs import user_config_dir
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from shillelagh import console


def test_main(mocker: MockerFixture) -> None:
    """
    Test ``main``.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = ["SELECT 1;", "", EOFError()]
    console.main()
    result = stdout.getvalue()
    assert result == "  1\n---\n  1\nGoodBye!\n"


def test_exception(mocker: MockerFixture) -> None:
    """
    Test that exceptions are captured and printed.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = ["SSELECT 1;", EOFError()]
    console.main()
    result = stdout.getvalue()
    assert result == 'SQLError: near "SSELECT": syntax error\nGoodBye!\n'


def test_ctrl_c(mocker: MockerFixture) -> None:
    """
    Test that ``CTRL-C`` does not exit the REPL.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [
        KeyboardInterrupt(),
        "SELECT 1;",
        EOFError(),
    ]
    console.main()
    result = stdout.getvalue()
    assert result == "  1\n---\n  1\nGoodBye!\n"


def test_configuration(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test loading the configuration file.
    """
    config_dir = Path(user_config_dir("shillelagh"))
    config_path = config_dir / "shillelagh.yaml"
    fs.create_file(config_path, contents=yaml.dump({"foo": {"bar": "baz"}}))

    connect = mocker.patch("shillelagh.console.connect")
    mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [EOFError()]
    console.main()

    connect.assert_called_with(":memory:", adapter_kwargs={"foo": {"bar": "baz"}})


def test_no_configuration(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test no configuration file found.
    """
    config_dir = Path(user_config_dir("shillelagh"))
    fs.create_dir(config_dir)

    connect = mocker.patch("shillelagh.console.connect")
    mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [EOFError()]
    console.main()

    connect.assert_called_with(":memory:", adapter_kwargs={})


def test_configuration_invalid(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test that an exception is raised if the configuration is invalid.
    """
    config_dir = Path(user_config_dir("shillelagh"))
    config_path = config_dir / "shillelagh.yaml"
    fs.create_file(config_path, contents="foo: *")

    _logger = mocker.patch("shillelagh.console._logger")
    mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [EOFError()]
    console.main()

    _logger.exception.assert_called_with("Unable to load configuration file")


def test_multiline(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test a simple multiline query
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = ["SELECT ", "1;", "", EOFError()]
    console.main()
    result = stdout.getvalue()
    assert result == "  1\n---\n  1\nGoodBye!\n"


def test_multiline_quoted_semicolon(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test a multiline query that contains quoted semicolons.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = ["SELECT ';'=", "';';", EOFError()]
    console.main()
    result = stdout.getvalue()

    assert result == "  ';'=\n   ';'\n------\n     1\nGoodBye!\n"


def test_multiline_quoted_semicolon_on_line_end(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test a multiline query that contains quoted semicolons on the line end.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = ["SELECT ';'=';", "';", EOFError()]
    console.main()
    result = stdout.getvalue()

    assert result == "  ';'=';\n       '\n--------\n       0\nGoodBye!\n"


def test_multiline_triple_quoted_semicolon_on_line_end(
    mocker: MockerFixture,
    fs: FakeFilesystem,
) -> None:
    """
    Test a multiline query that contains quoted semicolons on the line end.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [
        "SELECT ''';'''=''';",
        "''';",
        EOFError(),
    ]
    console.main()
    result = stdout.getvalue()

    assert (
        result
        == "  ''';'''=''';\n           '''\n--------------\n             0\nGoodBye!\n"
    )



"""
An adapter for in-memory Pandas dataframes.
"""

# pylint: disable=invalid-name

import inspect
import operator
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type

import numpy as np
import pandas as pd

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Boolean, DateTime, Field, Float, Integer, Order, String
from shillelagh.filters import (
    Equal,
    Filter,
    Impossible,
    IsNotNull,
    IsNull,
    NotEqual,
    Range,
)
from shillelagh.lib import SimpleCostModel
from shillelagh.typing import RequestedOrder, Row

# this is just a wild guess; used to estimate query cost
AVERAGE_NUMBER_OF_ROWS = 1000

type_map: Dict[str, Tuple[Type[Field], List[Type[Filter]]]] = {
    "i": (Integer, [Range, Equal, NotEqual, IsNull, IsNotNull]),
    "b": (Boolean, [Equal, NotEqual, IsNull, IsNotNull]),
    "u": (Integer, [Range, Equal, NotEqual, IsNull, IsNotNull]),
    "f": (Float, [Range, Equal, NotEqual, IsNull, IsNotNull]),
    "M": (DateTime, [Range, Equal, NotEqual, IsNull, IsNotNull]),
    "S": (String, [Range, Equal, NotEqual, IsNull, IsNotNull]),
    "O": (String, [Range, Equal, NotEqual, IsNull, IsNotNull]),
}


def get_field(dtype: np.dtype) -> Field:
    """
    Return a Shillelagh `Field` from a Numpy dtype.
    """
    class_, filters = type_map[dtype.kind]
    return class_(
        filters=filters,
        order=Order.ANY,
        exact=True,
    )


def find_dataframe(uri: str) -> Optional[pd.DataFrame]:
    """
    Go up the stack, find the Pandas dataframe.
    """
    for level in inspect.stack():
        # search locals
        context_locals = dict(inspect.getmembers(level[0]))["f_locals"]
        if uri in context_locals and isinstance(context_locals[uri], pd.DataFrame):
            return context_locals[uri]

        # search globals
        context_globals = dict(inspect.getmembers(level[0]))["f_globals"]
        if uri in context_globals and isinstance(context_globals[uri], pd.DataFrame):
            return context_globals[uri]

    return None


def get_df_data(  # pylint: disable=too-many-arguments, too-many-branches
    df: pd.DataFrame,
    columns: Dict[str, Field],
    bounds: Dict[str, Filter],
    order: List[Tuple[str, RequestedOrder]],
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> Iterator[Row]:
    """
    Apply the ``get_data`` method on a Pandas dataframe.
    """
    if df.empty:
        return

    # ensure column names are strings
    df = df.rename(columns={k: str(k) for k in df.columns})

    column_names = list(columns.keys())
    df = df[column_names]

    for column_name, filter_ in bounds.items():
        if isinstance(filter_, Impossible):
            return
        if isinstance(filter_, Equal):
            df = df[df[column_name] == filter_.value]
        elif isinstance(filter_, NotEqual):
            df = df[df[column_name] != filter_.value]
        elif isinstance(filter_, Range):
            if filter_.start is not None:
                operator_ = operator.ge if filter_.include_start else operator.gt
                df = df[operator_(df[column_name], filter_.start)]
            if filter_.end is not None:
                operator_ = operator.le if filter_.include_end else operator.lt
                df = df[operator_(df[column_name], filter_.end)]
        elif isinstance(filter_, IsNull):
            df = df[~df[column_name].notnull()]
        elif isinstance(filter_, IsNotNull):
            df = df[df[column_name].notnull()]
        else:
            raise ProgrammingError(f"Invalid filter: {filter_}")

    if order:
        by, requested_orders = list(zip(*order))
        ascending = [
            requested_order == Order.ASCENDING for requested_order in requested_orders
        ]
        df = df.sort_values(by=list(by), ascending=ascending)

    df = df[offset:]
    df = df[:limit]

    for row in df.itertuples(name=None):
        yield dict(zip(["rowid", *column_names], row))


def get_columns_from_df(df: pd.DataFrame) -> Dict[str, Field]:
    """
    Construct adapter columns from a Pandas dataframe.
    """
    return {
        # ensure column name is string
        str(column_name): get_field(dtype)
        for column_name, dtype in zip(df.columns, df.dtypes)
        if dtype.kind in type_map
    }


class PandasMemory(Adapter):

    """
    An adapter for in-memory Pandas dataframes.
    """

    safe = False

    supports_limit = True
    supports_offset = True

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        return find_dataframe(uri) is not None

    @staticmethod
    def parse_uri(uri: str) -> Tuple[str]:
        return (uri,)

    def __init__(self, uri: str):
        super().__init__()
        df = find_dataframe(uri)
        if df is None:
            raise ProgrammingError("Could not find dataframe")

        self.df = df
        self.columns = get_columns_from_df(df)

    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    get_cost = SimpleCostModel(AVERAGE_NUMBER_OF_ROWS)

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        yield from get_df_data(self.df, self.columns, bounds, order, limit, offset)

    def insert_data(self, row: Row) -> int:
        row_id: Optional[int] = row.pop("rowid")
        if row_id is None:
            row_id = max(self.df.index) + 1

        self.df.loc[row_id] = row

        return row_id

    def delete_data(self, row_id: int) -> None:
        self.df.drop([row_id], inplace=True)

    def update_data(self, row_id: int, row: Row) -> None:
        # the row_id might change on an update
        new_row_id = row.pop("rowid")
        if new_row_id != row_id:
            self.df.drop([row_id], inplace=True)

        self.df.loc[new_row_id] = row.values()



"""
Parse and format Google Sheet date/time patterns.

https://developers.google.com/sheets/api/guides/formats?hl=en#date_and_time_format_patterns
"""

# pylint: disable=invalid-name, fixme, broad-exception-raised

import calendar
import re
from collections import defaultdict
from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import Any, Dict, List, Tuple, Type, TypeVar, Union

from shillelagh.adapters.api.gsheets.parsing.base import LITERAL, Token, tokenize

DateTime = TypeVar("DateTime", datetime, date, time, timedelta)


class Meridiem(Enum):
    """
    Represent ante or post meridiem.
    """

    AM = "AM"
    PM = "PM"


class H(Token):
    """
    Hour of the day.

    Switches between 12 and 24 hour format depending on whether an am/pm
    indicator is present in the string.
    """

    regex = "h(?!h)"

    def format(self, value: Union[datetime, time], tokens: List[Token]) -> str:
        hour = value.hour

        if (
            any(token.__class__.__name__ in {"AP", "AMPM"} for token in tokens)
            and hour != 12
        ):
            hour %= 12

        # the 5th example in https://developers.google.com/sheets/api/guides/formats?hl=en
        # has a "PM" literal that switches to 12 hour format
        if (
            any(
                token.__class__.__name__ == "LITERAL"
                and ("AM" in token.token or "PM" in token.token)
                for token in tokens
            )
            and hour != 12
        ):
            hour %= 12

        return str(hour)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        match = re.match(r"\d{1,2}", value)
        if not match:
            raise Exception(f"Cannot parse value: {value}")
        size = len(match.group()) if 0 <= int(match.group()) < 24 else 1

        return {"hour": int(value[:size])}, value[size:]


class HHPlus(H):
    """
    Same as previous, but with a leading 0 for 1-9.
    """

    regex = "hh+"

    def format(self, value: Union[datetime, time], tokens: List[Token]) -> str:
        return super().format(value, tokens).zfill(2)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        return {"hour": int(value[:2])}, value[2:]


class M(Token):
    """
    If the previous non-literal token was hours or the subsequent one is
    seconds, then it represents minutes in the hour (no leading 0).
    Otherwise, it represents the month of the year as a number (no leading 0).
    """

    regex = "m(?!m)"

    def _is_minute(self, tokens: List[Token]) -> bool:
        """
        Return true if the token represents minutes, false if months.
        """
        is_minute = False
        i = -1
        for i, token in enumerate(tokens):
            if token is self:
                break
        else:
            raise Exception("Token is not present in list of tokens")

        for token in reversed(tokens[:i]):
            if token.__class__.__name__ == "LITERAL":
                continue
            if token.__class__.__name__ in {"H", "HHPlus"}:
                is_minute = True
            break

        for token in tokens[i + 1 :]:
            if token.__class__.__name__ == "LITERAL":
                continue
            if token.__class__.__name__ in {"S", "SS"}:
                is_minute = True
            break

        return is_minute

    def format(self, value: Union[date, datetime, time], tokens: List[Token]) -> str:
        if self._is_minute(tokens) and isinstance(value, (datetime, time)):
            return str(value.minute)

        if isinstance(value, (datetime, date)):
            return str(value.month)

        raise Exception(f"Cannot format value: {value}")

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        match = re.match(r"\d{1,2}", value)
        if not match:
            raise Exception(f"Cannot parse value: {value}")
        size = len(match.group()) if 1 <= int(match.group()) <= 24 else 1

        if self._is_minute(tokens):
            return {"minute": int(value[:size])}, value[size:]
        return {"month": int(value[:size])}, value[size:]


class MM(M):
    """
    As above, but with a leading 0 for both cases.
    """

    regex = "mm(?!m)"

    def format(self, value: Union[date, datetime, time], tokens: List[Token]) -> str:
        return super().format(value, tokens).zfill(2)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        if self._is_minute(tokens):
            return {"minute": int(value[:2])}, value[2:]
        return {"month": int(value[:2])}, value[2:]


class MMM(Token):
    """
    Three letter month abbreviation (e.g., "Feb").
    """

    regex = "mmm(?!m)"

    def format(self, value: Union[date, datetime, time], tokens: List[Token]) -> str:
        return value.strftime("%b")

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        month = datetime.strptime(value[:3], "%b").month
        return {"month": month}, value[3:]


class MMMM(MMM):
    """
    Full month name. mmmmmm+ also matches this.
    """

    regex = "(mmmm(?!m))|(m{6,})"

    def format(self, value: Union[date, datetime, time], tokens: List[Token]) -> str:
        return value.strftime("%B")

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        token = re.split(r"\b", value, 2)[1]
        size = len(token)
        month = datetime.strptime(value[:size], "%B").month
        return {"month": month}, value[size:]


class MMMMM(MMM):
    """
    First letter of the month (e.g., "J" for June).
    """

    regex = "mmmmm"

    def format(self, value: Union[date, datetime, time], tokens: List[Token]) -> str:
        return value.strftime("%B")[0]

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        letter = value[0]

        mapping = defaultdict(list)
        for i in range(1, 13):
            mapping[calendar.month_name[i][0]].append(i)
        if len(mapping[letter]) == 0:
            raise Exception(f"Unable to find month letter: {letter}")
        if len(mapping[letter]) > 1:
            raise Exception(f"Unable to parse month letter unambiguously: {letter}")

        return {"month": mapping[letter][0]}, value[1:]


class S(Token):
    """
    Seconds in the minute without a leading 0.
    """

    regex = "s(?!s)"

    def format(self, value: Union[datetime, time], tokens: List[Token]) -> str:
        return str(value.second)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        match = re.match(r"\d{1,2}", value)
        if not match:
            raise Exception(f"Cannot parse value: {value}")
        # leap seconds can be 60 or even 61
        size = len(match.group()) if 0 <= int(match.group()) <= 61 else 1

        return {"second": int(value[:size])}, value[size:]


class SS(S):
    """
    Seconds in the minute with a leading 0.
    """

    regex = "ss"

    def format(self, value: Union[datetime, time], tokens: List[Token]) -> str:
        return super().format(value, tokens).zfill(2)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        return {"second": int(value[:2])}, value[2:]


class HPlusDuration(Token):
    """
    Number of elapsed hours in a time duration. Number of letters indicates
    minimum number of digits (adds leading 0s).
    """

    regex = r"\[h+\]"

    def format(self, value: Union[timedelta], tokens: List[Token]) -> str:
        return str(int(value.total_seconds() // 3600)).zfill(len(self.token) - 2)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        size = len(self.token) - 2
        return {"hours": int(value[:size])}, value[size:]


class MPlusDuration(Token):
    """
    Number of elapsed minutes in a time duration. Number of letters indicates
    minimum number of digits (adds leading 0s).
    """

    regex = r"\[m+\]"

    def format(self, value: Union[timedelta], tokens: List[Token]) -> str:
        seconds = value.total_seconds()

        if any(token.__class__.__name__ == "HPlusDuration" for token in tokens):
            # ignore hours
            seconds %= 3600

        return str(int(seconds // 60)).zfill(len(self.token) - 2)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        size = len(self.token) - 2
        return {"minutes": int(value[:size])}, value[size:]


class SPlusDuration(Token):
    """
    Number of elapsed seconds in a time duration. Number of letters indicates
    minimum number of digits (adds leading 0s).
    """

    regex = r"\[s+\]"

    def format(self, value: Union[timedelta], tokens: List[Token]) -> str:
        seconds = value.total_seconds()

        if any(token.__class__.__name__ == "HPlusDuration" for token in tokens):
            # ignore hours
            seconds %= 3600

        if any(token.__class__.__name__ == "MPlusDuration" for token in tokens):
            # ignore minutes
            seconds %= 60

        return str(int(seconds)).zfill(len(self.token) - 2)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        size = len(self.token) - 2
        return {"seconds": int(value[:size])}, value[size:]


class D(Token):
    """
    Day of the month, no leading 0 for numbers less than 10.
    """

    regex = "d(?!d)"

    def format(self, value: Union[date, datetime], tokens: List[Token]) -> str:
        return str(value.day)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        match = re.match(r"\d{1,2}", value)
        if not match:
            raise Exception(f"Cannot parse value: {value}")
        size = len(match.group()) if 1 <= int(match.group()) <= 31 else 1
        return {"day": int(value[:size])}, value[size:]


class DD(D):
    """
    Day of the month, with a leading 0 for numbers less than 10.
    """

    regex = "dd(?!d)"

    def format(self, value: Union[date, datetime], tokens: List[Token]) -> str:
        return value.strftime("%d")

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        return {"day": int(value[:2])}, value[2:]


class DDD(D):
    """
    Day of the week, three letter abbreviation (e.g., "Mon").
    """

    regex = "ddd(?!d)"

    def format(self, value: Union[date, datetime], tokens: List[Token]) -> str:
        return value.strftime("%a")

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        return {"weekday": datetime.strptime(value[:3], "%a").weekday()}, value[3:]


class DDDDPlus(D):
    """
    Day of the week, full name.
    """

    regex = "d{4,}"

    def format(self, value: Union[date, datetime], tokens: List[Token]) -> str:
        return value.strftime("%A")

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        token = re.split(r"\b", value, 2)[1]
        size = len(token)
        return {"weekday": datetime.strptime(value[:size], "%A").weekday()}, value[
            size:
        ]


class YY(Token):
    """
    2-digit year.
    """

    regex = "y{1,2}(?!y)"

    def format(self, value: Union[date, datetime], tokens: List[Token]) -> str:
        return value.strftime("%y")

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        # assume 00 == 2000
        year = int(value[:2]) + 2000

        return {"year": year}, value[2:]


class YYYY(Token):
    """
    4-digit year.
    """

    regex = "y{3,}"

    def format(self, value: Union[date, datetime], tokens: List[Token]) -> str:
        return value.strftime("%Y")

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        return {"year": int(value[:4])}, value[4:]


class ZERO(Token):
    """
    Tenths of seconds. You can increase the precision to 2 digits
    with 00 or 3 digits (milliseconds) with 000.
    """

    regex = "0{1,3}(?!0)"

    def format(
        self,
        value: Union[datetime, time, timedelta],
        tokens: List[Token],
    ) -> str:
        precision = len(self.token)
        us = value.microseconds if isinstance(value, timedelta) else value.microsecond
        rounded = round(us / 1e6, precision)
        return str(int(rounded * 10**precision)).zfill(precision)

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        size = len(self.token)

        # adjust precision
        token = value[:size]
        token += "0" * (6 - size)
        microsecond = int(token)

        return {"microsecond": microsecond}, value[size:]


class AP(Token):
    """
    Displays "a" for AM, and "p" for PM. Also changes hours to 12-hour
    format. If the token letter is capitalized, the output is as well.
    """

    regex = "(a/p)|(A/P)"

    def format(self, value: Union[datetime, time], tokens: List[Token]) -> str:
        output = "a" if value.hour < 12 else "p"
        if self.token == "A/P":
            output = output.upper()
        return output

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        letter = value[:1]
        meridiem = Meridiem.PM if letter.upper() == "P" else Meridiem.AM
        return {"meridiem": meridiem}, value[1:]


class AMPM(AP):
    """
    As above, but displays "AM" or "PM" instead and is always capitalized.
    """

    regex = "am/pm"

    def format(self, value: Union[datetime, time], tokens: List[Token]) -> str:
        return "AM" if value.hour < 12 else "PM"

    def parse(self, value: str, tokens: List[Token]) -> Tuple[Dict[str, Any], str]:
        letter = value[:2]
        meridiem = Meridiem.PM if letter.upper() == "PM" else Meridiem.AM
        return {"meridiem": meridiem}, value[2:]


def parse_date_time_pattern(
    value: str,
    pattern: str,
    class_: Type[DateTime],
) -> DateTime:
    """
    Parse a value using a given pattern.

    See https://developers.google.com/sheets/api/guides/formats?hl=en.
    """
    classes = [
        H,
        HHPlus,
        M,
        MM,
        MMM,
        MMMM,
        MMMMM,
        S,
        SS,
        HPlusDuration,
        MPlusDuration,
        SPlusDuration,
        D,
        DD,
        DDD,
        DDDDPlus,
        YY,
        YYYY,
        AP,
        AMPM,
        ZERO,
        LITERAL,
    ]

    kwargs: Dict[str, Any] = {}
    tokens = list(tokenize(pattern, classes))
    for token in tokens:
        consumed, value = token.parse(value, tokens)
        kwargs.update(**consumed)

    # add PM offset
    if "hour" in kwargs:
        meridiem = kwargs.pop("meridiem", None)
        if meridiem == Meridiem.PM and kwargs["hour"] != 12:
            kwargs["hour"] += 12
        elif meridiem == Meridiem.AM and kwargs["hour"] == 12:
            kwargs["hour"] -= 12

    # we can't really do anything with ``weekday``
    if "weekday" in kwargs:
        del kwargs["weekday"]

    if "microsecond" in kwargs and class_ is timedelta:
        kwargs["microseconds"] = kwargs.pop("microsecond")

    try:
        return class_(**kwargs)
    except TypeError as ex:
        raise Exception("Unsupported format") from ex


def format_date_time_pattern(value: DateTime, pattern: str) -> str:
    """
    Format a date/time related object to a given pattern.

    See https://developers.google.com/sheets/api/guides/formats?hl=en.
    """
    classes = [
        H,
        HHPlus,
        M,
        MM,
        MMM,
        MMMM,
        MMMMM,
        S,
        SS,
        HPlusDuration,
        MPlusDuration,
        SPlusDuration,
        D,
        DD,
        DDD,
        DDDDPlus,
        YY,
        YYYY,
        AP,
        AMPM,
        ZERO,
        LITERAL,
    ]

    parts = []
    tokens = list(tokenize(pattern, classes))
    for token in tokens:
        parts.append(token.format(value, tokens))

    return "".join(parts)



"""
Tests for the system adapter.
"""
from datetime import datetime, timezone
from unittest import mock

from freezegun import freeze_time
from pytest_mock import MockerFixture

from shillelagh.adapters.api.system import SystemAPI
from shillelagh.backends.apsw.db import connect


def test_system_cpu(mocker: MockerFixture) -> None:
    """
    Test a simple CPU query.
    """
    psutil = mocker.patch("shillelagh.adapters.api.system.psutil")
    psutil.cpu_count.return_value = 4
    psutil.cpu_percent.side_effect = [
        [1, 2, 3, 4],
        [1, 2, 3, 4],
        [1, 2, 3, 4],
        [1, 2, 3, 4],
    ]

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT * FROM "system://cpu"
        LIMIT 2
    """
    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(cursor.execute(sql))
    assert data == [
        (datetime(2021, 1, 1, tzinfo=timezone.utc), 0.01, 0.02, 0.03, 0.04),
        (datetime(2021, 1, 1, tzinfo=timezone.utc), 0.01, 0.02, 0.03, 0.04),
    ]

    psutil.cpu_percent.assert_has_calls(
        [
            mock.call(interval=1, percpu=True),
            mock.call(interval=1, percpu=True),
        ],
    )


def test_system_memory(mocker: MockerFixture) -> None:
    """
    Test a simple memory query.
    """
    mocker.patch("shillelagh.adapters.api.system.time")
    psutil = mocker.patch("shillelagh.adapters.api.system.psutil")
    psutil.virtual_memory()._asdict.return_value = {
        "total": 34359738368,
        "available": 15130095616,
        "percent": 56.0,
        "used": 18285113344,
        "free": 1579941888,
        "active": 13551853568,
        "inactive": 13460545536,
        "wired": 4733259776,
    }

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT * FROM "system://memory"
        LIMIT 1
    """
    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(cursor.execute(sql))
    assert data == [
        (
            datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            34359738368,
            15130095616,
            56.0,
            18285113344,
            1579941888,
            13551853568,
            13460545536,
            4733259776,
        ),
    ]


def test_system_swap(mocker: MockerFixture) -> None:
    """
    Test a simple swap memory query.
    """
    mocker.patch("shillelagh.adapters.api.system.time")
    psutil = mocker.patch("shillelagh.adapters.api.system.psutil")
    psutil.swap_memory()._asdict.return_value = {
        "total": 18253611008,
        "used": 16865034240,
        "free": 1388576768,
        "percent": 92.4,
        "sin": 1010873262080,
        "sout": 4259106816,
    }

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT * FROM "system://swap"
        LIMIT 1
    """
    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(cursor.execute(sql))
    assert data == [
        (
            datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            18253611008,
            16865034240,
            1388576768,
            92.4,
            1010873262080,
            4259106816,
        ),
    ]


def test_system_different_interval(mocker: MockerFixture) -> None:
    """
    Test a simple CPU query with a custom interval.
    """
    psutil = mocker.patch("shillelagh.adapters.api.system.psutil")
    psutil.cpu_count.return_value = 4
    psutil.cpu_percent.side_effect = [
        [1, 2, 3, 4],
        [1, 2, 3, 4],
        [1, 2, 3, 4],
        [1, 2, 3, 4],
    ]

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT * FROM "system://cpu?interval=2"
        LIMIT 2
    """
    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(cursor.execute(sql))
    assert data == [
        (datetime(2021, 1, 1, tzinfo=timezone.utc), 0.01, 0.02, 0.03, 0.04),
        (datetime(2021, 1, 1, tzinfo=timezone.utc), 0.01, 0.02, 0.03, 0.04),
    ]

    psutil.cpu_percent.assert_has_calls(
        [
            mock.call(interval=2, percpu=True),
            mock.call(interval=2, percpu=True),
        ],
    )


def test_system_interrupt(mocker: MockerFixture) -> None:
    """
    Test CPU query interrupt.
    """
    psutil = mocker.patch("shillelagh.adapters.api.system.psutil")
    psutil.cpu_count.return_value = 4
    psutil.cpu_percent.side_effect = [
        [1, 2, 3, 4],
        [1, 2, 3, 4],
        KeyboardInterrupt(),
    ]

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT * FROM "system://cpu"
    """
    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(cursor.execute(sql))
    assert data == [
        (datetime(2021, 1, 1, tzinfo=timezone.utc), 0.01, 0.02, 0.03, 0.04),
        (datetime(2021, 1, 1, tzinfo=timezone.utc), 0.01, 0.02, 0.03, 0.04),
    ]


def test_get_data_cpu(mocker: MockerFixture) -> None:
    """
    Test ``get_data`` with the ``CPU`` resource.
    """
    psutil = mocker.patch("shillelagh.adapters.api.system.psutil")
    psutil.cpu_count.return_value = 4
    psutil.cpu_percent.side_effect = [
        [1, 2, 3, 4],
        [1, 2, 3, 4],
        [1, 2, 3, 4],
        [1, 2, 3, 4],
    ]
    time = mocker.patch("shillelagh.adapters.api.system.time")

    adapter = SystemAPI("cpu")

    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(adapter.get_data({}, [], limit=2, offset=1))
    assert data == [
        {
            "rowid": 0,
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "cpu0": 0.01,
            "cpu1": 0.02,
            "cpu2": 0.03,
            "cpu3": 0.04,
        },
        {
            "rowid": 1,
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "cpu0": 0.01,
            "cpu1": 0.02,
            "cpu2": 0.03,
            "cpu3": 0.04,
        },
    ]

    time.sleep.assert_called_with(1.0)

    data = list(adapter.get_data({}, [], limit=2, offset=1, requested_columns={"cpu1"}))
    assert data == [
        {"rowid": 0, "cpu1": 0.02},
        {"rowid": 1, "cpu1": 0.02},
    ]


def test_get_data_all(mocker: MockerFixture) -> None:
    """
    Test ``get_data`` with the ``ALL`` resource.
    """
    psutil = mocker.patch("shillelagh.adapters.api.system.psutil")
    psutil.cpu_count.return_value = 4
    psutil.cpu_percent.return_value = [1, 2, 3, 4]
    psutil.virtual_memory()._asdict.return_value = {
        "total": 34359738368,
        "available": 15130095616,
        "percent": 56.0,
        "used": 18285113344,
        "free": 1579941888,
        "active": 13551853568,
        "inactive": 13460545536,
        "wired": 4733259776,
    }
    psutil.swap_memory()._asdict.return_value = {
        "total": 18253611008,
        "used": 16865034240,
        "free": 1388576768,
        "percent": 92.4,
        "sin": 1010873262080,
        "sout": 4259106816,
    }
    time = mocker.patch("shillelagh.adapters.api.system.time")

    adapter = SystemAPI("all")

    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(adapter.get_data({}, [], limit=1))
    assert data == [
        {
            "rowid": 0,
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "cpu0": 0.01,
            "cpu1": 0.02,
            "cpu2": 0.03,
            "cpu3": 0.04,
            "virtual_total": 34359738368,
            "virtual_available": 15130095616,
            "virtual_percent": 56.0,
            "virtual_used": 18285113344,
            "virtual_free": 1579941888,
            "virtual_active": 13551853568,
            "virtual_inactive": 13460545536,
            "virtual_wired": 4733259776,
            "swap_total": 18253611008,
            "swap_used": 16865034240,
            "swap_free": 1388576768,
            "swap_percent": 92.4,
            "swap_sin": 1010873262080,
            "swap_sout": 4259106816,
        },
    ]

    time.sleep.assert_not_called()

    psutil.cpu_percent.reset_mock()
    psutil.virtual_memory.reset_mock()
    psutil.swap_memory.reset_mock()

    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(
            adapter.get_data(
                {},
                [],
                limit=1,
                requested_columns={"virtual_percent"},
            ),
        )
    assert data == [{"rowid": 0, "virtual_percent": 56.0}]

    psutil.cpu_percent.assert_not_called()
    psutil.virtual_memory.assert_called()
    psutil.swap_memory.assert_not_called()

    time.sleep.assert_called_with(1.0)


def test_get_data_bogus(mocker: MockerFixture) -> None:
    """
    Test ``get_data`` with a bogus ``requested_columns``.
    """
    mocker.patch("shillelagh.adapters.api.system.time")

    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(
            SystemAPI("all").get_data({}, [], limit=1, requested_columns={"bogus"}),
        )
    assert data == [{"rowid": 0}]

    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(
            SystemAPI("cpu").get_data({}, [], limit=1, requested_columns={"bogus"}),
        )
    assert data == [{"rowid": 0}]

    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(
            SystemAPI("memory").get_data({}, [], limit=1, requested_columns={"bogus"}),
        )
    assert data == [{"rowid": 0}]

    with freeze_time("2021-01-01T00:00:00Z"):
        data = list(
            SystemAPI("swap").get_data({}, [], limit=1, requested_columns={"bogus"}),
        )
    assert data == [{"rowid": 0}]



"""
Test the HTML table scraper.
"""
import pandas as pd
import pytest
from pytest_mock import MockerFixture

from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError


def test_html_table(mocker: MockerFixture) -> None:
    """
    Test basic operations with a dataframe.
    """
    df = pd.DataFrame(  # noqa: F841  pylint: disable=unused-variable, invalid-name
        [
            {"index": 10, "temperature": 15.2, "site": "Diamond_St"},
            {"index": 11, "temperature": 13.1, "site": "Blacktail_Loop"},
            {"index": 12, "temperature": 13.3, "site": "Platinum_St"},
            {"index": 13, "temperature": 12.1, "site": "Kodiak_Trail"},
        ],
    )
    mock_pd = mocker.patch("shillelagh.adapters.api.html_table.pd")
    mock_pd.read_html.return_value = [df]

    connection = connect(":memory:", adapters=["htmltableapi"])
    cursor = connection.cursor()
    sql = 'SELECT * FROM "https://example.org/"'
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]

    mock_pd.read_html.side_effect = Exception("Invalid URL")

    connection = connect(":memory:", adapters=["htmltableapi"])
    cursor = connection.cursor()
    sql = 'SELECT * FROM "https://example.org/"'
    with pytest.raises(ProgrammingError) as excinfo:
        cursor.execute(sql)
    assert str(excinfo.value) == "Unsupported table: https://example.org/"


def test_html_table_fragment(mocker: MockerFixture) -> None:
    """
    Test the fragment used for the table index.
    """
    df = pd.DataFrame(  # noqa: F841  pylint: disable=unused-variable, invalid-name
        [
            {"index": 10, "temperature": 15.2, "site": "Diamond_St"},
            {"index": 11, "temperature": 13.1, "site": "Blacktail_Loop"},
            {"index": 12, "temperature": 13.3, "site": "Platinum_St"},
            {"index": 13, "temperature": 12.1, "site": "Kodiak_Trail"},
        ],
    )
    mock_pd = mocker.patch("shillelagh.adapters.api.html_table.pd")
    mock_pd.read_html.return_value = [None, df]

    connection = connect(":memory:", adapters=["htmltableapi"])
    cursor = connection.cursor()
    sql = 'SELECT * FROM "https://example.org/#1"'
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]

    mock_pd.read_html.return_value = [df]

    connection = connect(":memory:", adapters=["htmltableapi"])
    cursor = connection.cursor()
    sql = 'SELECT * FROM "https://example.org/#anchor"'
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]
